from json import dump
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Iterator

import pytest

from sugarloaf_utilities.template import GenerateTemplate


@pytest.fixture
def output_dir() -> Iterator[Path]:
    with TemporaryDirectory() as output_dir:
        output_dir = Path(output_dir)

        sugarloaf_config_dir = output_dir / ".sugarloaf"
        sugarloaf_config_dir.mkdir()
        with open(sugarloaf_config_dir / "config.json", "w") as file:
            dump(dict(name="test"), file)

        yield output_dir

@pytest.fixture
def override_dir(output_dir) -> Path:
    override_path = output_dir / "override"
    override_path.mkdir()
    return override_path

def test_simple_merge(output_dir, override_dir):
    """
    Ensure that we merge template files and overrides
    """
    app_setup_path = override_dir / "app-setup"
    app_setup_path.mkdir()

    with open(app_setup_path / "test.tf", "w") as file:
        file.write("TEST")

    template_generator = GenerateTemplate()
    template_generator(output_dir, override_dir)

    build_directory = output_dir / template_generator.build_path

    # Templates were transferred over
    assert (build_directory / "app-monitoring").exists()

    # User overrides were transferred over
    assert (build_directory / "app-setup/test.tf").exists()


def test_terraform_state_file_in_output(output_dir, override_dir):
    """
    Ensure that we don't delete terraform state files that live within the build path
    """
    template_generator = GenerateTemplate()
    monitoring_directory = output_dir / template_generator.build_path / "app-monitoring"
    monitoring_directory.mkdir(parents=True)

    with open(monitoring_directory / "terraform.tfstate", "w") as file:
        file.write("STATE")

    # Perform the generation
    template_generator(output_dir, override_dir, confirm_delete=False)

    with open(monitoring_directory / "terraform.tfstate") as file:
        file.read().strip() == "STATE"



from dataclasses import dataclass, field, replace
from datetime import datetime
from json import dumps as json_dumps, loads as json_loads
from re import match
from time import sleep, time
from typing import List, Optional, Set

from click import command, option, secho
from kubernetes import client, watch
from urllib3.exceptions import TimeoutError


class DeploymentError(Exception):
    pass


@dataclass
class ParsedDockerImage:
    """
    Component of a docker image URL. There are a few different forms that these
    URLs can take:
    - postgres:latest
    - postgres
    - docker.pkg.dev/postgres:latest

    """
    image: str

    repository: Optional[str] = None
    tag: Optional[str] = None

    @classmethod
    def from_url(cls, url):
        # optional parse repository, force parse image, optional tag
        parsed_image = match(r"(.*\/)?([a-zA-Z0-9-]*)(:[a-zA-Z0-9-]+)?", url)
        repository = parsed_image.group(1)
        tag = parsed_image.group(3)

        return cls(
            repository=repository.strip("/") if repository else None,
            image=parsed_image.group(2),
            tag=tag.strip(":") if tag else None,
        )

    @property
    def url(self):
        """
        Return the full docker image URL
        """
        url = ""
        url += f"{self.repository}/" if self.repository else ""
        url += self.image
        url += f":{self.tag}" if self.tag else ""
        return url


@dataclass
class UpdatedPod:
    name: str
    ready: bool = False


@dataclass
class UpdatedDeployment:
    """
    Capture the metadata around a deployment that was updated as part
    of this revision process

    """
    namespace: str
    name: str
    # `app` label used by pods to group themselves
    app_name: str
    # index of the previously/currently deployed replicaset
    previous_revision: int
    current_revision: int
    # the replica name that backs the current revision
    current_replica_name: str
    current_pods: Optional[List[UpdatedPod]] = None

    @property
    def is_stabilized(self) -> bool:
        """
        :param: true if all pod deployments have stabilized

        """
        # Client callers have not yet populated the pods, therefore we don't
        # know whether we have a stable deployment
        if self.current_pods is None:
            return False
        if not self.current_pods:
            return True
        return all(pod.ready for pod in self.current_pods)


@dataclass(frozen=True)
class UniqueContainer:
    """
    Uniquely defines a container, by image and environment variables.

    """
    namespace: str
    image: str

    env_serialized: str
    env_from_serialized: str

    env: List[client.V1EnvVar] = field(compare=False)
    env_from: List[client.V1EnvFromSource] = field(compare=False)
    deployment_name: str = field(compare=False)

    @classmethod
    def from_container(cls, deployment, container):
        api_client = client.ApiClient()
        container_env = api_client.sanitize_for_serialization(container.env)
        container_env_from = api_client.sanitize_for_serialization(container.env_from)

        return cls(
            namespace=deployment.metadata.namespace,
            image=container.image,
            env=container.env,
            env_from=container.env_from,
            env_serialized=cls.serialize_env(container_env),
            env_from_serialized=cls.serialize_env(container_env_from),
            deployment_name=deployment.metadata.name,
        )

    @staticmethod
    def serialize_env(obj):
        def _sort(child):
            if isinstance(child, dict):
                return sorted((k, _sort(v)) for k, v in child.items())
            if isinstance(child, list):
                return sorted(_sort(x) for x in child)
            else:
                return child
        return json_dumps(_sort(obj))


REVERSION_KEY = "deployment.kubernetes.io/revision"
CHANGE_CAUSE_KEY = "kubernetes.io/change-cause"


class DeploymentManager:
    def __init__(self, stabilization_limit: int, max_migration_limit: int, should_migrate: bool):
        self.apps_api = client.AppsV1Api()
        self.core_api = client.CoreV1Api()
        self.batch_api = client.BatchV1Api()
        self.stabilization_limit = stabilization_limit
        self.max_migration_limit = max_migration_limit
        self.should_migrate = should_migrate

    def __call__(self, image: str, new_sha: str, commit_description: Optional[str] = None):
        if self.should_migrate:
            self.run_migration(image=image, new_sha=new_sha)

        updated_deployments = self.update_deployments_with_image(image=image, new_sha=new_sha, commit_description=commit_description)
        failed_stabilization = self.get_failed_stabilization(updated_deployments)

        if failed_stabilization:
            # Now rollback the changes that didn't succeed
            secho("Will rollback changes", fg="red")
            self.rollback_deployments(failed_stabilization)
            secho("Rollback complete", fg="green")

            # Report failure at the CI level
            raise DeploymentError()

        secho("All deployments succeeded", fg="green")

    def update_deployments_with_image(self, image, new_sha, commit_description) -> List[UpdatedDeployment]:
        """
        Update deployments that reference the given image - updating the sha and pushing
        a new revision into the cluster.

        """
        updated_deployments: List[UpdatedDeployment] = []

        for item, containers in self.find_deployments_with_image(image):
            for container in containers:
                parsed = ParsedDockerImage.from_url(container.image)

                previous_revision = int(item.metadata.annotations.get(REVERSION_KEY))
                app_name = item.spec.template.metadata.labels.get("app")

                secho(f"Updating {item.metadata.name}", fg="yellow")

                new_image = replace(parsed, tag=new_sha)
                container.image = new_image.url

            # Limit the current git commit description to the first 200 chars to make it more readable in
            # the change logs
            if commit_description:
                item.metadata.annotations[CHANGE_CAUSE_KEY] = self.get_trimmed_string(commit_description, 200)

            # kubernetes.io/change-cause=
            self.apps_api.patch_namespaced_deployment(
                name=item.metadata.name,
                namespace=item.metadata.namespace,
                body=item,
            )

            # get the latest revision
            deployment = self.apps_api.read_namespaced_deployment(item.metadata.name, item.metadata.namespace)
            current_revision = int(deployment.metadata.annotations.get(REVERSION_KEY))

            # If it's the same as the previous one, we didn't actually update anything
            # (which is unusual but we should handle this case gracefully)
            if previous_revision == current_revision:
                secho(f"Revisions equal for `{item.metadata.name}` - skipping stabilization check", fg="yellow")
                continue

            new_deployed_replica = self.get_replica_for_revision(app_name, current_revision)
            new_replica_name = new_deployed_replica.metadata.name

            updated_deployments.append(
                UpdatedDeployment(
                    namespace=item.metadata.namespace,
                    name=item.metadata.name,
                    app_name=app_name,
                    previous_revision=previous_revision,
                    current_revision=current_revision,
                    current_replica_name=new_replica_name,
                )
            )

        return updated_deployments

    def get_failed_stabilization(self, updated_deployments):
        """
        :returns deployments that could not reach a healthy state:
        """
        start_timer = time()

        while (
            pending_stabilization := [
                deployment
                for deployment in updated_deployments
                if not deployment.is_stabilized
            ]
        ):
            if time() - start_timer > self.stabilization_limit:
                formatted_failures = "\n".join([f"- {current_deployment.name} [{current_deployment.current_pods}]" for current_deployment in pending_stabilization])
                secho(f"Failed to stabilize deployments:\n{formatted_failures}", fg="red")

                # Log more information about each run
                for current_deployment in pending_stabilization:
                    for pod in current_deployment.current_pods:
                        if pod.ready:
                            continue
                        # By default will only return the most recent pod's failure in the case of restarts
                        # To modify this behavior we can pass `previous=True`
                        logs = self.core_api.read_namespaced_pod_log(name=pod.name, namespace=current_deployment.namespace, pretty="true")
                        secho(f"-- Logs: {pod.name} --", fg="red")
                        print(logs)
                        secho(f"-- End Logs: {pod.name}  --", fg="red")
                break

            for updated_deployment in pending_stabilization:
                updated_deployment.current_pods = []

                pods = self.core_api.list_namespaced_pod(
                    namespace=updated_deployment.namespace,
                    label_selector=f"app={updated_deployment.app_name}"
                )
                for pod in pods.items:
                    is_new_replica = any(
                        [
                            owner.name == updated_deployment.current_replica_name
                            for owner in pod.metadata.owner_references
                        ]
                    )

                    # This pod is part of the new rollout group, which we care about monitoring
                    # for stability. Get the current status.
                    if is_new_replica:
                        is_ready = all([
                            status.ready for status in (pod.status.container_statuses or [])
                        ])
                        updated_deployment.current_pods.append(
                            UpdatedPod(
                                name=pod.metadata.name,
                                ready=is_ready,
                            )
                        )

                # If all are ready we are stable
                if updated_deployment.is_stabilized:
                    secho(f"Deployment has stabilized: `{updated_deployment.name}`", fg="green")

            sleep(1)

        return pending_stabilization

    def rollback_deployments(self, rollback_deployments):
        for updated_deployment in rollback_deployments:
            deployment = self.apps_api.read_namespaced_deployment(updated_deployment.name, updated_deployment.namespace)
            matching_set = self.get_replica_for_revision(updated_deployment.app_name, updated_deployment.previous_revision)

            # Update the spec
            deployment.spec.template = matching_set.spec.template

            secho(f"Bumping revision `{updated_deployment.namespace}/{updated_deployment.name}`: {updated_deployment.current_revision} -> {updated_deployment.current_revision+1}", fg="yellow")
            self.apps_api.patch_namespaced_deployment(
                name=updated_deployment.name,
                namespace=updated_deployment.namespace,
                body=client.V1Deployment(
                    spec=deployment.spec,
                    metadata=client.V1ObjectMeta(
                        annotations={REVERSION_KEY: str(updated_deployment.current_revision+1)}
                    )
                ),
            )

    def run_migration(self, image: str, new_sha: str):
        """
        Run the application's `migrate` entrypoint, if available, before deploying the services.

        Specifically we will run the migration command on any unique containers that are found
        that match our `image` search term. We will inject the same unique environment variables
        into the pod definitions.

        To avoid race conditions in these migration scripts, we only run one at a time.

        """
        secho("Running optional migration", fg="yellow")

        # Combination of (container, env variables)
        unique_containers: Set[UniqueContainer] = set()
        for deployment, containers in self.find_deployments_with_image(image):
            for container in containers:
                unique_container = UniqueContainer.from_container(deployment, container)

                parsed = ParsedDockerImage.from_url(unique_container.image)
                new_image = replace(parsed, tag=new_sha)
    
                unique_containers.add(
                    replace(unique_container, image=new_image.url)
                )

        # Create a job to migrate each of these containers
        for container in unique_containers:
            migration_name = f"migration-{container.deployment_name}-{time()}"
            secho(f" Launch: `{migration_name}`")
            job = self.batch_api.create_namespaced_job(
                namespace=container.namespace,
                body=client.V1Job(
                    metadata=client.V1ObjectMeta(
                        name=migration_name,
                    ),
                    spec=client.V1JobSpec(
                        completions=1,
                        template=client.V1PodTemplateSpec(
                            metadata=client.V1ObjectMeta(
                                labels={
                                    "job-type": "migration",
                                }
                            ),
                            spec=client.V1PodSpec(
                                containers=[
                                    client.V1Container(
                                        name="migration",
                                        image=container.image,
                                        args=["migrate"],
                                        env=container.env,
                                        env_from=container.env_from,
                                    )
                                ],
                                restart_policy="OnFailure",
                            )
                        )
                    )
                )
            )

            w = watch.Watch()

            try:
                for event in w.stream(self.batch_api.list_namespaced_job, namespace=job.metadata.namespace, _request_timeout=self.max_migration_limit):
                    if event["object"].metadata.name == job.metadata.name:
                        if event["object"].status.succeeded:
                            secho(" Migration succeeded", fg="green")
                            break
            except TimeoutError:
                # Find the pod name that was spawned as part of this job
                # for logging purposes
                migration_pods = self.core_api.list_namespaced_pod(namespace=container.namespace, label_selector=f"job-type={job.metadata.name}")
                migration_pod_names = [pod.metadata.name for pod in migration_pods.items]

                secho(f"Migration timed out; check process logs: {', '.join(migration_pod_names)}", fg="red")
                raise TimeoutError("Migration timed out")

    def find_deployments_with_image(self, image: str):
        """
        Find all deployments with the given app name.

        :returns: an iterator of (deployment, [matching containers])

        """
        result = self.apps_api.list_deployment_for_all_namespaces()

        for item in result.items:
            matches_search = []
            for container in item.spec.template.spec.containers:
                parsed = ParsedDockerImage.from_url(container.image)
                if parsed.image == image:
                    matches_search.append(container)
            if matches_search:
                yield item, matches_search

    def get_replica_for_revision(self, app_name: str, revision: int):
        """
        Return the replicaset definition for a given app revision
        """
        replica_sets = self.apps_api.list_replica_set_for_all_namespaces(
            label_selector=f"app={app_name}",
        )

        # Find the one that was present right before our upgrade
        matching_set = [
            replica_set
            for replica_set in replica_sets.items
            if int(replica_set.metadata.annotations.get(REVERSION_KEY)) == int(revision)
        ]

        if not matching_set:
            raise ValueError("No matching revision")

        return matching_set[0]

    @staticmethod
    def get_trimmed_string(string, max_length):
        if len(string) > max_length:
            return string[:max_length] + "..."
        return string


@command()
@option("--image", required=True)
@option("--new-sha", required=True)
@option("--commit-description")
@option("--stabilization-limit", default=30)
@option("--migration-limit", default=240)
@option("--migrate", is_flag=True, default=True)
def update_image(image, new_sha, commit_description, stabilization_limit, migration_limit, migrate):
    """
    Scans through currently deployed definitions and determines which ones contain
    references to the `image` artifact and therefore need to be updated with the
    given sha.

    :param stabilization_limit: The seconds to wait for the deployment health to stabilize
    :param migration_limit: The seconds to wait for the migration to complete

    """
    manager = DeploymentManager(stabilization_limit, migration_limit, should_migrate=migrate)
    manager(image, new_sha, commit_description)



from re import M
from unittest.mock import patch

import pytest
from kubernetes import client

from sugarloaf_utilities.deployment import REVERSION_KEY, DeploymentError, DeploymentManager


class GenerationUtilities:
    def __init__(
        self,
        app_name: str,
        image_name: str,
        container_name: str,
    ):
        self.app_name = app_name
        self.image_name = image_name
        self.container_name = container_name

    def generate_pod_template(self):
        return client.V1PodTemplateSpec(
            metadata=client.V1ObjectMeta(
                labels={
                    "app": self.app_name,
                }
            ),
            spec=client.V1PodSpec(
                containers=[
                    client.V1Container(
                        name=self.container_name,
                        image=self.image_name,
                    )
                ]
            )
        )

    def generate_deployment_spec(self, revision: int):
        return client.V1Deployment(
            metadata=client.V1ObjectMeta(
                name="test-deplyoment",
                namespace="default",
                annotations={
                    REVERSION_KEY: str(revision),
                }
            ),
            spec=client.V1DeploymentSpec(
                template=self.generate_pod_template(),
                selector=client.V1LabelSelector(
                    match_labels={
                        "app": self.app_name,
                    }
                )
            )
        )

    def generate_replica_spec(self, replica_set_name: str, revision: int):
        return client.V1ReplicaSet(
            metadata=client.V1ObjectMeta(
                name=replica_set_name,
                annotations={
                    REVERSION_KEY: str(revision),
                }
            ),
            spec=client.V1ReplicaSetSpec(
                replicas=1,
                selector=client.V1LabelSelector(
                    match_labels={
                        "app": self.app_name,
                    }
                ),
                template=self.generate_pod_template(),
            )
        )

    def generate_pod_spec(self, owner_name: str, ready: bool):
        return client.V1Pod(
            metadata=client.V1ObjectMeta(
                name="test-pod",
                namespace="default",
                labels={
                    "app": self.app_name,
                },
                owner_references=[
                    client.V1OwnerReference(
                        api_version="apps/v1",
                        kind="ReplicaSet",
                        name=owner_name,
                        uid=f"{owner_name}-uid",
                    )
                ]
            ),
            status=client.V1PodStatus(
                container_statuses=[
                    client.V1ContainerStatus(
                        name=self.container_name,
                        ready=ready,
                        image=self.image_name,
                        image_id=f"{self.image_name}-id",
                        restart_count=0,
                    )
                ]
            )
        )

def test_clean_deploy():
    """
    Deploy new payload and pods become healthy immediatey
    """
    manager = DeploymentManager(5, 0, should_migrate=False)

    image_name="test-image"
    replica_set_name = "test-replica-set"
    spec_generator = GenerationUtilities(
        app_name="test-app",
        image_name=image_name,
        container_name="test-container",
    )

    with (
        patch.object(manager.apps_api, "list_deployment_for_all_namespaces") as list_deployment_mock,
        patch.object(manager.apps_api, "patch_namespaced_deployment") as patch_deployment_mock,
        patch.object(manager.apps_api, "read_namespaced_deployment") as read_namespaced_deployment_mock,
        patch.object(manager.apps_api, "list_replica_set_for_all_namespaces") as list_replica_set_for_all_namespaces_mock,
        patch.object(manager.core_api, "list_namespaced_pod") as list_namespaced_pod_mock,
    ):
        # List an instance of the deployment that has our container image included
        list_deployment_mock.return_value = client.V1DeploymentList(
            items=[
                spec_generator.generate_deployment_spec(revision=1),
            ]
        )

        # No-op for the patch deployment itself
        patch_deployment_mock.return_value = dict()

        # After deploy, should bump the revision number
        read_namespaced_deployment_mock.return_value = spec_generator.generate_deployment_spec(revision=2)

        # Find the replica set that backs the new deployment
        list_replica_set_for_all_namespaces_mock.return_value = client.V1ReplicaSetList(
            items=[
                spec_generator.generate_replica_spec(replica_set_name, revision=2),
            ]
        )

        # Pods are ready immediately
        list_namespaced_pod_mock.return_value = client.V1PodList(
            items=[
                spec_generator.generate_pod_spec(replica_set_name, ready=True),
            ]
        )

        manager(image_name, "new-tag")

def test_erroring_deploy():
    """
    Deploy new payload and pods stay unhealthy
    """
    manager = DeploymentManager(5, 0, should_migrate=False)

    image_name="test-image"
    replica_set_name = "test-replica-set"
    spec_generator = GenerationUtilities(
        app_name="test-app",
        image_name=image_name,
        container_name="test-container",
    )

    with (
        patch.object(manager.apps_api, "list_deployment_for_all_namespaces") as list_deployment_mock,
        patch.object(manager.apps_api, "patch_namespaced_deployment") as patch_deployment_mock,
        patch.object(manager.apps_api, "read_namespaced_deployment") as read_namespaced_deployment_mock,
        patch.object(manager.apps_api, "list_replica_set_for_all_namespaces") as list_replica_set_for_all_namespaces_mock,
        patch.object(manager.core_api, "list_namespaced_pod") as list_namespaced_pod_mock,
        patch.object(manager.core_api, "read_namespaced_pod_log") as read_namespaced_pod_log_mock,
    ):
        # List an instance of the deployment that has our container image included
        list_deployment_mock.return_value = client.V1DeploymentList(
            items=[
                spec_generator.generate_deployment_spec(revision=1),
            ]
        )

        # No-op for the patch deployment itself
        patch_deployment_mock.return_value = dict()

        # After deploy, should bump the revision number
        read_namespaced_deployment_mock.return_value = spec_generator.generate_deployment_spec(revision=2)

        # Find the replica set that backs both deployments, since we'll have to revert
        # to the old spec
        list_replica_set_for_all_namespaces_mock.return_value = client.V1ReplicaSetList(
            items=[
                spec_generator.generate_replica_spec(replica_set_name, revision=1),
                spec_generator.generate_replica_spec(replica_set_name, revision=2),
            ]
        )

        # Pods are ready immediately
        list_namespaced_pod_mock.return_value = client.V1PodList(
            items=[
                spec_generator.generate_pod_spec(replica_set_name, ready=False),
            ]
        )

        # Show no logs
        read_namespaced_pod_log_mock.return_value = ""

        with pytest.raises(DeploymentError):
            manager(image_name, "new-tag")



from time import sleep

from click import command, option, secho, Choice, group
from kubernetes import client


@command()
@option("--metric-name", required=True)
def get_external_metric_value(metric_name):
    api = client.CustomObjectsApi()
    resource = api.list_namespaced_custom_object(group="external.metrics.k8s.io", version="v1beta1", namespace="default", plural=metric_name)
    for item in resource["items"]:
        secho(f"Value: {item['value']} - {item['timestamp']}")

@group()
def bastion():
    """
    Grouper of bastion / jump-host utilities that allow us to access resources tied to kubernetes
    while still respecting VPC permissions.

    """
    pass

@bastion.command()
@option("--namespace", default="default")
@option(
    "--image",
    type=Choice(["google/cloud-sdk", "ubuntu", "postgres"]),
    required=True
)
def launch_cli_pod(namespace, image):
    """
    Launch a simple pod with the google cloud CLI pre-installed in order
    to engage with the cluster in a remote shell.

    If you want to test artifact repository permissions, do one of:
    ```
    gcloud auth list
    gcloud auth configure-docker us-central1-docker.pkg.dev
    docker manifest inspect {package}
    ```

    """
    name = "test-cli-pod"
    body = {
        "apiVersion": "v1",
        "kind": "Pod",
        "metadata": {
            "name": name,
        },
        "spec": {
            "containers": [
                {
                    #"image": "google/cloud-sdk:latest",
                    "image": f"{image}:latest",
                    "name": name,
                    "command": ["sleep","infinity"]
                }
            ],
            **(
                {
                    "nodeSelector": {
                        "iam.gke.io/gke-metadata-server-enabled": "true"
                    }
                }
                if image == "google/cloud-sdk"
                else {}
            )
        }
    }

    api = client.CoreV1Api()
    api.create_namespaced_pod(body=body, namespace=namespace)

    try:
        secho(f"Launched pod.\nAccess with: `kubectl exec -it {name} --namespace {namespace} -- /bin/bash`", fg="green")

        # TODO: Add check for current status, pending or running

        while True:
            sleep(1)
    except KeyboardInterrupt:
        secho("Shutting down gracefully...")

        api.delete_namespaced_pod(
            namespace=namespace,
            name=name,
        )


@bastion.command()
@option("--namespace", default="default")
@option("--host", required=True)
@option("--database", required=True)
@option("--username", required=True)
@option("--local-port", required=True)
def access_database(namespace, host, database, username, local_port):
    """
    Proxy to the remote database to debug with local psql.

    We faciliate this by spawning a connection pooler into a remote namespace, which
    we expect has VPC firewall access to connect to the database.

    All CLI parameters should be for the RDS database that you wish to access.

    """
    secho("Enter the database password:")
    password = input(" > ")
    if not password:
        raise ValueError("No password specified.")

    name = "database-proxy-pod"
    config_name = f"{name}-config"

    env = {
        "DB_HOST": host,
        "DB_NAME": database,
        "DB_USER": username,
        "DB_PASSWORD": password,
        "AUTH_TYPE": "scram-sha-256"
    }

    # Create a configmap to store the user config file
    # We need to do this manually because the default userlist.txt file uses
    # md5 password encryption. The latest postgres uses scram-sha which is
    # incompatible.
    config_body = {
        "metadata": {
            "name": config_name,
        },
        "data": {
            "userlist.txt": f'"{username}" "{password}"'
        }
    }

    body = {
        "apiVersion": "v1",
        "kind": "Pod",
        "metadata": {
            "name": name,
        },
        "spec": {
            "containers": [
                {
                    "image": f"edoburu/pgbouncer:latest",
                    "name": name,
                    "env": [
                        {
                            "name": key,
                            "value": value,
                        }
                        for key, value in env.items()
                    ],
                    "port": {
                        "container_port": 5432
                    },
                    "volumeMounts": [
                        {
                            "name": "config-volume",
                            "mountPath": "/etc/pgbouncer/userlist.txt",
                            "subPath": "userlist.txt"
                        }
                    ]
                }
            ],
            "volumes": [
                {
                    "name": "config-volume",
                    "configMap": {
                        "name": config_name,
                    }
                }
            ]
        }
    }

    api = client.CoreV1Api()

    api.create_namespaced_config_map(body=config_body, namespace=namespace)
    api.create_namespaced_pod(body=body, namespace=namespace)

    try:
        secho(f"Launched pod.\nAccess with: `kubectl port-forward {name} --namespace {namespace} {local_port}:5432`", fg="green")

        # TODO: Add check for current status, pending or running

        while True:
            sleep(1)
    except KeyboardInterrupt:
        secho("Shutting down gracefully...")

        api.delete_namespaced_config_map(
            namespace=namespace,
            name=config_name,
        )

        api.delete_namespaced_pod(
            namespace=namespace,
            name=name,
        )



from sugarloaf_utilities.terraform import should_init
import pytest

@pytest.mark.parametrize("message,expected", [
    # Sometimes errors are broken per line
    ('This module is not yet installed. Run\n"terraform init" to install all\nmodules required by this configuration.', True)
])
def test_parse_init_error(message, expected):
    assert should_init(message) == expected



from multiprocessing.sharedctypes import Value
from os import getcwd
from pathlib import Path
from subprocess import run
from typing import List, Optional

from click import (
    Choice as ClickChoice,
    Path as ClickPath,
    command,
    option,
    secho,
)

from sugarloaf_utilities.components import get_ordered_packages
from sugarloaf_utilities.io import EchoSubprocess
from sugarloaf_utilities.template import GenerateTemplate


def find_parent_config_path(child_path: str) -> Path:
    # Traverse up the filepath until we find a .sugarloaf folder
    # that contains a configuration file
    child_path = Path(child_path)
    while child_path is not None:
        if (child_path / ".sugarloaf").exists():
            return child_path
        child_path = child_path.parent
    raise ValueError("No .sugarloaf configuration folder found in any parent directories.")


def should_init(message) -> bool:
    """
    Given the `stderr` response from a terraform execution, attempt to determine
    whether any of these errors are init related. If so we should re-initialize the module.

    """
    # If we see these errors during apply, it means we need to re-init our directory
    init_errors = [
        "Required plugins are not installed",
        "Module not installed",
        "Inconsistent dependency lock file",
        "The source address was changed",
        # Fallback - terraform usually hints to update the files
        "Run \"terraform init\""
    ]

    return any(
        error_text in message.replace("\n", " ")
        for error_text in init_errors
    )

def run_chained_terraform(apply_path, command: str, subset: Optional[List[str]], auto_approve: bool):
    packages = get_ordered_packages()
    package_names = {package.name for package in packages}

    if subset:
        # Determine if we were provided allowable entries
        missing_subset = [item for item in subset if item not in package_names]
        if missing_subset:
            secho(f"{missing_subset} are not valid packages, only supported: {package_names}", fg="red")
            return
        packages = [package for package in packages if package.name in subset]

    missing_packages = []
    for package in packages:
        if not (apply_path / package.name).exists():
            missing_packages.append(package.name)
    if missing_packages:
        secho(f"Packages not found: {missing_packages}", fg="red")
        return

    for package in packages:
        secho(f"Executing package: `{package.name}`...", fg="yellow")

        # Do the short setup, assuming we have already inited the parent container.
        # If we catch any errors that have to do with initializing we fallback to running the init
        response_apply = EchoSubprocess(["terraform", f"-chdir={(apply_path / package.name)}", command, "-auto-approve" if auto_approve else ""]).run()
        if response_apply.returncode != 0:
            # Attempt to see whether any of these errors are init related
            if not should_init(response_apply.stderr):
                # NOTE: Encountering this error might mean we need to expand our `INIT_ERRORS` constant with
                # further error codes that are related to non-inited packages
                # Since we rely on string matching we know our approach is a bit fuzzy
                secho(f"Plan failed for module `{package.name}`, permanent error. Aborting...", fg="red")
                return
        else:
            secho("Success", fg="green")
            continue

        # Do the full setup
        secho(f"Running init and apply for `{package.name}`", fg="yellow")
        response_init = EchoSubprocess(["terraform", f"-chdir={(apply_path / package.name)}", "init"]).run()
        if response_init.returncode != 0:
            secho(f"Init failed for module `{package.name}`, aborting...", fg="red")
            return
        response_apply = run(["terraform", f"-chdir={(apply_path / package.name)}", "apply", "-auto-approve" if auto_approve else ""])
        if response_apply.returncode != 0:
            secho(f"Plan failed for module `{package.name}`, aborting...", fg="red")
            return
        secho("Success", fg="green")


@command()
@option("--path", type=ClickPath(exists=True, file_okay=False, dir_okay=True), required=False)
@option("--subset", type=str, multiple=True, required=False)
@option("--debug", is_flag=True, default=False)
@option("-auto-approve", is_flag=True, default=False)
def chain_apply(path: Optional[str], subset: Optional[List[str]], auto_approve: bool, debug: bool):
    """
    Chained apply that "terraform apply" our packages in dependency order, by default requesting
    clients confirm each change that is applied. To override and have it accept changes by default, pass
    the -auto-approve flag. We will halt on any plan failure before attempting to run the next ones in the
    dependency graph.

    Will attempt to apply in the local directory. If all packages specified in the DAG definition
    are not found, will throw an error.

    :param debug: If this flag is provided, will refresh the template during every run. Useful for
    local debugging of the `sugarloaf-infrastructure` codebase.

    """
    current_path = Path(path) if path else Path(getcwd())

    if not (current_path / ".sugarloaf/config.json").exists():
        secho("Must be run in root project directory", fg="red")
        return

    # Generate new build configuration based on the latest core files
    if debug:
        #project_root = find_parent_config_path(current_path)
        secho("Debug: generating new templates...\n", fg="yellow")
        template_generator = GenerateTemplate()
        template_generator(current_path, current_path / "override", confirm_delete=False)
        secho("\n")

    apply_path = current_path / "build-infrastructure"

    run_chained_terraform(apply_path, "apply", subset, auto_approve)


@command()
@option("--path", type=ClickPath(exists=True, file_okay=False, dir_okay=True), required=False)
@option("--subset", type=ClickChoice([package.name for package in get_ordered_packages()]), multiple=True, required=False)
@option("-auto-approve", is_flag=True, default=False)
def chain_destroy(path: Optional[str], subset: Optional[List[str]], auto_approve: bool):
    """
    Chained destroy, will tear down resources in the opposite order that they are created. Accepts
    the same parameters as chain_apply.

    """
    current_path = Path(path) if path else Path(getcwd())

    if not (current_path / ".sugarloaf/config.json").exists():
        secho("Must be run in root project directory", fg="red")
        return

    apply_path = current_path / "build-infrastructure"

    run_chained_terraform(apply_path, "destroy", subset, auto_approve)



from click import group
from kubernetes import config

from sugarloaf_utilities.debug import bastion
from sugarloaf_utilities.deployment import update_image
from sugarloaf_utilities.secrets import secrets
from sugarloaf_utilities.terraform import chain_apply, chain_destroy


config.load_kube_config()

@group()
def main():
    pass

main.add_command(update_image)
main.add_command(bastion)
main.add_command(chain_apply, name="apply")
main.add_command(chain_destroy, name="destroy")
main.add_command(secrets)



from multiprocessing.sharedctypes import Value
from os import link, remove, symlink
from pathlib import Path
from pkg_resources import resource_filename
from re import compile as re_compile
from shutil import copyfile, rmtree
from typing import Optional, Union

from click import (
    Path as ClickPath,
    command,
    option,
    secho,
)

from sugarloaf_utilities.components import get_ordered_packages


def get_terraform_path(asset_name: Optional[str] = None) -> Path:
    submodule_path = Path("terraform")
    if asset_name:
        submodule_path /= asset_name
    return Path(resource_filename(__name__, str(submodule_path)))


class GenerateTemplate:
    def __init__(self, verbose=False):
        self.source_module_path = get_terraform_path()
        self.package_names = {package.name for package in get_ordered_packages()}
        self.verbose = verbose

        # Files that often appear on disk but aren't intended for copy
        self.filename_blacklist = {".DS_Store"}

        # The compiled template will be built into this folder
        self.build_path = "build-infrastructure"

        # Keep the following files names during a overwrite of existing infrastructure
        # This allows clients to upgrade in a slightly smoother way without losing terraform state
        self.keep_files_in_migrate = [
            re_compile(r".*\.terraform.*"),
            re_compile(r".*\.tfstate"),
            re_compile(r".*\.tfstate\..*"),
            re_compile(r".*\.tflock"),
            re_compile(r".*\.tflock\..*"),
            re_compile(r".*\.terraform.lock"),
            re_compile(r".*\.terraform.lock\..*"),
        ]

    def __call__(self, output_path: Union[Path, str], override_path: Union[Path, str], confirm_delete: bool = True):
        output_path = Path(output_path)
        override_path = Path(override_path)

        if confirm_delete and not self.confirm_folder_ready(output_path):
            return

        output_infrastructure_path = output_path / self.build_path
        self.copy_base_modules(output_infrastructure_path)

        if override_path.exists():
            self.copy_user_overrides(output_infrastructure_path, override_path)

    def confirm_folder_ready(self, output_path: Union[Path, str]):
        """
        Determine whether we have already populated the output directory with
        our module files - if so, challenge and ask for user confirmation.

        """
        output_path = Path(output_path)

        if not (output_path / ".sugarloaf/config.json").exists():
            secho("Could not find the `.sugarloaf/config.json` file in this path.", fg="red")
            return False

        has_existing_data = any(
            (output_path / name).exists()
            for name in self.package_names
        )

        if not has_existing_data:
            return True

        secho("Existing infrastructure modules found.", fg="red")
        secho("Type the name of the output folder to continue...")
        confirmation = input("  > ")
        if confirmation != output_path.name:
            secho("Confirmation failed.", fg="red")
            return False
        return True

    def copy_base_modules(self, output_path: Path):
        """
        Copy the template files that are from the centralized infrastructure repo

        """

        for package_name in self.package_names:
            source_path = (self.source_module_path / package_name)
            if not source_path.exists():
                raise ValueError(f"Unable to find source package at path: `{source_path}`.\nYou might need to relink paths with `link-template`.")

            # Delete the destination path since we have received user confirmation
            # This cleans up any old files that may have been left over from a previous run, even
            # if we have removed their contents in a newer revision
            if (output_path / package_name).exists():
                self.remove_old_template_files(output_path / package_name)

            secho(f"Creating package: {package_name}", fg="green")

            for file_source_path in source_path.glob("**/*"):
                if not file_source_path.is_file() or self.file_in_blacklist(file_source_path):
                    continue
                file_output_path = output_path / file_source_path.relative_to(source_path.parent)
                # Create the parent directories if necessary first
                file_output_path.parent.mkdir(parents=True, exist_ok=True)
                copyfile(file_source_path, file_output_path)

    def copy_user_overrides(self, output_path: Path, override_path: Path):
        """
        Copy user-override files according to our conventions:
        - folders are supported as-is
        - the first root folder must be one of our predefined core packages

        """
        for path in override_path.glob("**/*"):
            if not path.is_file():
                continue
            if self.file_in_blacklist(path):
                secho(f"Skipping user-overide file, in blacklist: `{path}`", fg="yellow")
                continue

            destination_folder = str(path.relative_to(override_path)).split("/")[0]
            if destination_folder not in self.package_names:
                raise ValueError(f"Destination folder name is invalid: {path.name}")

            relative_path = path.relative_to(override_path)
            secho(f"Copying custom file: {relative_path}", fg="green")

            file_output = output_path / relative_path
            file_output.parent.mkdir(parents=True, exist_ok=True)

            # File already exists as a base template, we should append the user configuration
            # to the file contents
            if file_output.exists():
                secho(f"Duplicate user override & template file found `{file_output}`, appending to existing file.", fg="yellow")
                secho("This will unlink the build files from their original source, which might not be desirable.", fg="yellow")

                with open(file_output) as file:
                    existing_content = file.read()
                with open(path) as file:
                    new_content = file.read()

                remove(file_output)

                with open(file_output, "w") as file:
                    file.write(f"#\n# Auto Merge: Template\n#\n{existing_content}\n#\n# Auto Merge: User Override\n#\n{new_content}")
            else:
                link(path, file_output)

    def remove_old_template_files(self, root_path: Union[Path, str]):
        """
        Attempts to intelligently remove old template & override files since the values
        might have changed on disk and we don't want to end up with old files.

        """
        root_path = Path(root_path)

        # Clear out the files first according to our rules
        for delete_path in root_path.glob("**/*"):
            if not delete_path.is_file():
                continue
            should_remove = not any(
                keep_pattern.match(str(delete_path))
                for keep_pattern in self.keep_files_in_migrate
            )
            if should_remove:
                if self.verbose:
                    secho(f"Will remove outdated: `{delete_path.relative_to(root_path)}`")
                remove(delete_path)

        # Clear out any folders that are now empty
        for delete_path in self.walk_directories_dfs(root_path):
            if not delete_path.iterdir():
                remove(delete_path)

    def file_in_blacklist(self, path: Union[Path, str]) -> str:
        """
        Extends blacklist support to folders as well; ie. if any parts of the file path
        match the blacklist, we must

        """
        # also blacklist any parent folders
        return len(set(str(path).split("/")) & self.filename_blacklist) > 0

    def walk_directories_dfs(self, folder_path: Union[Path, str]):
        """
        Walk folder in dfs order, ie. returns most specific folders first. This method
        is useful to do a conditional deletion if folders have nested empty folders.

        """
        for path in Path(folder_path).iterdir(): 
            if path.is_dir(): 
                yield from self.walk_directories_dfs(path)
                continue
            yield path

@command()
@option("--output-path", type=ClickPath(exists=True, file_okay=False, dir_okay=True), required=True)
@option("--verbose", type=bool, default=False)
def main(output_path, verbose):
    """
    Template variables given the definition folder.

    We allow an "override" folder within the destination path. This allows clients to place custom
    terraform files into our centralized folders (ie. core-management/custom.tf, app-setup/custom.tfvars, etc)
    that are copied when the other files are re-generated. By convention, name your override
    files `{folder-name}/{file-name}.tf`. So for the previous examples you want `override/core-management/custom.tf`.
    We also support more complicated nesting structure like: `override/core-management/testing/custom.tf`.
    We create a hard link vs a symlink between these files because we want present as a full file to local
    utilities yet guarantee that if one file is edited the other is as well.

    """
    output_path = Path(output_path)

    template_generator = GenerateTemplate(verbose=verbose)
    template_generator(output_path, output_path / "override")


@command()
@option("--template-path", type=ClickPath(exists=True, file_okay=False, dir_okay=True), required=True)
@option("--hard", is_flag=True, default=False)
def link_templates(template_path: str, hard: bool):
    """
    Link the core template path (in the root of the git repostory) with symbolic links
    to the templates folder of the python package. This allows us to ship these files as
    part of the manifest to pypi.

    :param hard: Whether to hard-link the individual template files. This is necessary before
        deploying the package because sdist can only navigate paths that appear physically within
        the python package filetree. By default (when hard is not provided) it will soft-link the
        files, which is preferrable for local development. This soft linking allows the package to
        read underlying changes to the core modules while modifying locally.

    """
    source_path = Path(template_path).expanduser().resolve()

    desintation_path = get_terraform_path()
    if desintation_path.exists():
        rmtree(desintation_path)
    desintation_path.mkdir()

    for package in get_ordered_packages():
        source_package_path = source_path / package.name
        destination_package_path = desintation_path / package.name

        if not source_package_path.exists():
            raise ValueError(f"Source package does not exist: `{source_package_path}`")

        secho(f"Link paths: `{source_package_path}`->`{destination_package_path}`", fg="green")
        if hard:
            # Trace through each individual file and link these
            secho(" - Link: hard")
            for copy_file_path in source_package_path.glob("**/*"):
                if not copy_file_path.is_file():
                    continue
                relative_file_path = copy_file_path.relative_to(source_package_path)
                destination_file_path = destination_package_path / relative_file_path
                destination_file_path.parent.mkdir(parents=True, exist_ok=True)

                link(copy_file_path, destination_file_path)
        else:
            # Link the whole folder
            secho(" - Link: soft")
            symlink(source_package_path, destination_package_path)



from asyncio import (
    TimeoutError,
    create_subprocess_shell,
    run as async_run,
    wait_for,
)
from asyncio.subprocess import PIPE
from dataclasses import dataclass
from typing import List


@dataclass
class SubprocessResult:
    stdout: str
    stderr: str
    returncode: int


class EchoSubprocess:
    """
    Echos a subprocess in realtime and returns the result after its complete. This can
    be thought of as combining Popen's capture_output and a streaming output to the shared
    parent process's stdout. Also allows for stdin input during execution.

    """
    def __init__(self, command: List[str]):
        self.command = " ".join(command)

    def run(self):
        """
        Spawn the subprocess and return the execution log.
        """
        return async_run(self.echo_subprocess_async())

    async def echo_subprocess_async(self, wait_interval=5):
        p = await create_subprocess_shell(self.command, stdout=PIPE, stderr=PIPE)

        logs = {
            "stdout": "",
            "stderr": "",
        }

        while True:
            for log_key, stream in [("stdout", p.stdout), ("stderr", p.stderr)]:
                # read data byte-by-byte to consume all pending characters in the buffer
                # requesting a full newline (via readline) will timeout without returning
                # if a newline hasn't yet been echoed
                while True:
                    # Get up to the buffer size and immediately return
                    new_content = await stream.read(1024)

                    print(new_content.decode(errors="ignore"), end="", flush=True)
                    logs[log_key] += new_content.decode(errors="ignore")

                    if not new_content:
                        break

            if p.returncode is not None:
                break

        return SubprocessResult(
            stdout="".join(logs["stdout"]),
            stderr="".join(logs["stderr"]),
            returncode=p.returncode
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sugarloaf_utilities', 'sugarloaf_utilities.tests']

package_data = \
{'': ['*']}

install_requires = \
['click>=7.0,<8.0', 'kubernetes>=24.2.0,<25.0.0', 'rich>=12.5.0,<13.0.0']

entry_points = \
{'console_scripts': ['create-infrastructure = '
                     'sugarloaf_utilities.template:main',
                     'infra = sugarloaf_utilities.main:main',
                     'link-templates = '
                     'sugarloaf_utilities.template:link_templates']}

setup_kwargs = {
    'name': 'sugarloaf-utilities',
    'version': '0.1.2',
    'description': '',
    'long_description': None,
    'author': 'Pierce Freeman',
    'author_email': 'pierce@freeman.vc',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



from dataclasses import dataclass, field
from graphlib import TopologicalSorter
from typing import List
from uuid import UUID, uuid4


@dataclass
class InfrastructurePackage:
    """
    Defines a terraform package that should be applied at runtime. The dependency_on
    property builds up a dependency graph of dependencies and applies them such to maximize
    parallelism while respecting the order involved. This allows clients to refactor their
    global terraform packages to be more efficient at the package level (isolation of components
    to decrease terrform plan speeds with roundtrip fetch of resource state).

    """
    name: str
    depends_on: List["InfrastructurePackage"] = field(default_factory=list)

    # Internal identifier
    identifier: UUID = field(default_factory=uuid4)


core_infrastructure = InfrastructurePackage(
    name="core-infrastructure",
)

core_definitions = InfrastructurePackage(
    name="core-definitions",
    depends_on=[core_infrastructure],
)

app_monitoring = InfrastructurePackage(
    name="app-monitoring",
    depends_on=[core_definitions],
)

app_setup = InfrastructurePackage(
    name="app-setup",
    depends_on=[
        # relies on service monitoring definitions to be completed
        core_definitions,
        # relies on the monitoring namespace to already be setup
        app_monitoring
    ],
)

def get_ordered_packages() -> List[InfrastructurePackage]:
    """
    Resolves the DAG Dependency by sniffing the global space for InfrastructurePackages instances.
    This avoids us having to specify a primary list of the packages.

    """
    packages = [item for item in globals().values() if isinstance(item, InfrastructurePackage)]
    package_by_id = {
        package.identifier: package
        for package in packages
    }

    edges = {
        package.identifier: [dependency.identifier for dependency in package.depends_on]
        for package in packages
    }

    sorted = TopologicalSorter(edges)
    return [
        package_by_id[identifier]
        for identifier in sorted.static_order()
    ]
